import json
import urllib.request


def normalize_package_name(name: str):
    return name.lower().replace("_", "-")


def get_pkg_latest(name: str) -> str:
    with urllib.request.urlopen(
        f"https://pypi.python.org/pypi/{name}/json", timeout=30
    ) as conn:
        version_data = json.load(conn)
    return str(version_data["info"]["version"])



from __future__ import annotations

import argparse
import collections
import difflib
import os
import pathlib
import sys
import typing as t

from . import config, yaml
from .package_utils import normalize_package_name


def load_precommit_config() -> tuple[dict[str, t.Any], None | str | tuple[str, ...]]:
    path = pathlib.Path.cwd() / ".pre-commit-config.yaml"
    if not path.is_file():
        raise ValueError("upadup cannot run without .pre-commit-config.yaml")

    with path.open() as fp:
        return yaml.load(fp), fp.newlines


def update_dependency(current_dependency, known_dependency_names, dependency_versions):
    if "==" not in current_dependency:
        return None

    package_name, _, old_version = current_dependency.partition("==")

    normed_pkg = normalize_package_name(package_name)
    if normed_pkg not in known_dependency_names:
        return None

    new_version = dependency_versions[normed_pkg]
    if old_version == new_version:
        return None

    return f"{package_name}=={dependency_versions[normed_pkg]}"


def build_updated_dependency_map(
    hook_config, known_dependency_names, dependency_versions
):
    new_deps = {}
    for current in hook_config.get("additional_dependencies", ()):
        new_dependency = update_dependency(
            current, known_dependency_names, dependency_versions
        )
        if new_dependency is None:
            continue
        new_deps[current] = new_dependency
    return new_deps


def _sort_updates_key(update):
    current_dependency, new_dependency = update
    return (current_dependency.lc.line, current_dependency.lc.col)


def generate_updates(hook_config, additional_dependencies, dependency_versions):
    print(
        f"upadup is checking additional_dependencies of {hook_config['id']}...", end=""
    )
    new_deps = build_updated_dependency_map(
        hook_config, additional_dependencies, dependency_versions
    )
    if new_deps:
        print()
        for current_dependency, new_dependency in new_deps.items():
            print(f"  {current_dependency} => {new_dependency}")
            yield (current_dependency, new_dependency)
    else:
        print("no updates needed")


def create_new_content(config_path: pathlib.Path, updates):
    with config_path.open("r") as fp:
        file_content = fp.readlines()

    # NB: int() == 0
    line_offsets: dict[int, int] = collections.defaultdict(int)
    for old_dep, new_dep in updates:
        lineno, column = old_dep.lc.line, old_dep.lc.col

        begin = column + line_offsets[lineno]
        end = begin + len(old_dep)
        line_offsets[lineno] += len(new_dep) - len(old_dep)

        old_line = file_content[lineno]
        file_content[lineno] = "".join((old_line[:begin], new_dep, old_line[end:]))

    return file_content


def generate_diff(config_path: pathlib.Path, updates):
    new_content = create_new_content(config_path, updates)
    with config_path.open("r") as fp:
        old_content = fp.readlines()
    return difflib.unified_diff(
        old_content, new_content, ".pre-commit-config.yaml", ".pre-commit-config.yaml"
    )


def apply_updates(
    config_path: pathlib.Path, updates, newline: None | str | tuple[str, ...]
):
    file_content = create_new_content(config_path, updates)

    # If no newlines were encountered, use the OS default.
    if newline is None:
        newline = os.linesep
    # If multiple newline variants were encountered, pick one.
    # Note that the order of newlines in the tuple is meaningless.
    if isinstance(newline, tuple):
        newline = newline[0]

    with config_path.open("w", newline=newline) as fp:
        fp.write("".join(file_content))


def main(argv: list[str] | None = None):
    parser = argparse.ArgumentParser(
        description="upadup -- the pre-commit additional_dependencies updater"
    )
    parser.add_argument(
        "--check",
        help="check and show diff, but do not update",
        action="store_true",
        default=False,
    )
    args = parser.parse_args(argv or sys.argv[1:])

    upadup_config = config.load_upadup_config()
    precommit_config, existing_newlines = load_precommit_config()

    all_updates = []
    for repo_config in precommit_config["repos"]:
        repo_str = repo_config.get("repo").casefold()
        # Strip the ".git" suffix from the repo URL, if present.
        if repo_str.endswith(".git"):
            repo_str = repo_str[:-4]
        if repo_str in upadup_config["repos"]:
            upadup_repo_config = upadup_config["repos"][repo_str]
            for hook_config in repo_config["hooks"]:
                hook_id = hook_config["id"]
                if hook_id in upadup_repo_config:
                    all_updates.extend(
                        generate_updates(
                            hook_config,
                            upadup_repo_config.get(hook_id, []),
                            upadup_config["versions"],
                        )
                    )

    all_updates = sorted(all_updates, key=_sort_updates_key)

    if all_updates:
        if args.check:
            print(
                "".join(
                    generate_diff(
                        pathlib.Path.cwd() / ".pre-commit-config.yaml", all_updates
                    )
                )
            )
            sys.exit(1)
        else:
            print("apply updates...", end="")
            apply_updates(
                pathlib.Path.cwd() / ".pre-commit-config.yaml",
                all_updates,
                existing_newlines,
            )
            print("done")
    else:
        print("no updates needed in any hook configs")



from __future__ import annotations

import pathlib
import typing as t

from . import yaml
from .package_utils import get_pkg_latest, normalize_package_name


class BadConfigError(ValueError):
    def __init__(self, message: str) -> None:
        super().__init__(f"malformed config: {message}")


DEFAULT_CONFIG = yaml.load(
    """\
repos:
  - repo: https://github.com/pycqa/flake8
    hooks:
      - id: flake8
        additional_dependencies:
          - flake8-bandit
          - flake8-bugbear
          - flake8-builtins
          - flake8-comprehensions
          - flake8-docstrings
          - flake8-implicit-str-concat
          - flake8-logging-format
          - flake8-pyi
          - flake8-typing-as-t
          - flake8-typing-imports

  - repo: https://github.com/adamchainz/blacken-docs
    hooks:
      - id: blacken-docs
        additional_dependencies:
          - black

  # Old hook URLs
  # -------------

  - repo: https://github.com/asottile/blacken-docs
    hooks:
      - id: blacken-docs
        additional_dependencies:
          - black
"""
)


def _read_upadup_config() -> dict[str, t.Any]:
    path = pathlib.Path.cwd() / ".upadup.yaml"
    if not path.is_file():
        return {}
    with path.open() as fp:
        return yaml.load(fp)


def _validate_config(conf_dict: dict[str, t.Any]) -> None:
    if "repos" in conf_dict:
        if not isinstance(conf_dict["repos"], list):
            raise BadConfigError("'$.repos' should be list")
        for i_repo, repo_config in enumerate(conf_dict["repos"]):
            if not isinstance(repo_config, dict):
                raise BadConfigError(f"'$.repos[{i_repo}]' should be map")
            if not isinstance(repo_config.get("repo"), str):
                raise BadConfigError(f"'$.repos[{i_repo}].repo' should be string")
            if not isinstance(repo_config.get("hooks"), list):
                raise BadConfigError(f"'$.repos[{i_repo}].hooks' should be list")
            for i_hook, hook_config in enumerate(repo_config["hooks"]):
                if not isinstance(hook_config["id"], str):
                    raise BadConfigError(
                        f"'$.repos[{i_repo}].hooks[{i_hook}].id' should be string"
                    )
                if not isinstance(hook_config.get("additional_dependencies"), list):
                    raise BadConfigError(
                        f"'$.repos[{i_repo}].hooks[{i_hook}].additional_dependencies' "
                        "should be list"
                    )
                for i_ad, dep in enumerate(hook_config["additional_dependencies"]):
                    if not isinstance(dep, str):
                        raise BadConfigError(
                            f"'$.repos[{i_repo}].hooks[{i_hook}]"
                            f".additional_dependencies[{i_ad}]' should be string"
                        )

    if "extends_default" in conf_dict:
        if not isinstance(conf_dict["extends_default"], bool):
            raise ValueError("malformed config, '$.extends_default' should be boolean")


def _populate_map(
    conf_map: dict[str, t.Any], full_config: dict[str, t.Any], versions: dict[str, str]
) -> None:
    for repo_config in full_config.get("repos", []):
        repo_str = repo_config["repo"].casefold()
        if repo_str not in conf_map:
            conf_map[repo_str] = {}
        for hook_config in repo_config["hooks"]:
            hook_id = hook_config["id"]
            additional_dependencies = hook_config["additional_dependencies"]

            for dep in [normalize_package_name(n) for n in additional_dependencies]:
                if dep not in versions:
                    versions[dep] = get_pkg_latest(dep)

            if hook_id not in conf_map[repo_str]:
                conf_map[repo_str][hook_id] = additional_dependencies
            else:
                conf_map[repo_str][hook_id].extend(additional_dependencies)


def load_upadup_config() -> dict[str, t.Any]:
    upadup_config = _read_upadup_config()
    _validate_config(upadup_config)

    versions: dict[str, str] = {}
    upadup_config_map: dict[str, t.Any] = {}

    if upadup_config.get("extends_default", True):
        _populate_map(upadup_config_map, DEFAULT_CONFIG, versions)
    _populate_map(upadup_config_map, upadup_config, versions)

    return {"repos": upadup_config_map, "versions": versions}
